""" Library of Satellite Constellations """

import numpy as np
import sgp4.api
import astropy.time
import astropy.units
import astropy.coordinates
import matplotlib.pyplot as plt

class Constellation:
    """
    Base class for all constellations
    """
    
    def get_teme_position(self, time):
        """
        Gets positions of all satellites in TEME coordinate frame

        :param time: Time at which to get positions
        :type time: :class:`astropy.time.Time`
        :return: (x, y, z) position measured in meters
        :rtype: tuple[:class:`np.ndarray]
        """

        jd1 = np.array([time.jd1])
        jd2 = np.array([time.jd2])

        _, teme_pos, _ = self.constellation.sgp4(jd1, jd2)
        teme_pos = teme_pos[:,0,:]
        x, y, z = teme_pos[:,0], teme_pos[:,1], teme_pos[:,2]

        return x * 1000, y * 1000, z * 1000
    
    def get_hcs_position(self, time, earth_location):
        """
        Gets position of all satellites in constellation in HCS frame
        
        :param time: Time at which to get positions
        :type time: :class:`astropy.time.Time`
        :param earth_location: Location at which to get positions in HCS frame
        :type earth_location: :class:`astropy.coordinates.EarthLocation`
        :return: altitudes (degrees), azimuths (degrees), heights (meters)
        :rtype: tuple[:class:`np.ndarray`]
        """

        x, y, z = self.get_teme_position(time)
        teme = astropy.coordinates.TEME(
            x = x * astropy.units.meter,
            y = y * astropy.units.meter,
            z = z * astropy.units.meter,
            representation_type = 'cartesian',
            obstime = time)
        
        itrs_geo = teme.transform_to(astropy.coordinates.ITRS(obstime = time))
        topo_itrs_repr = itrs_geo.cartesian.without_differentials() \
                        - earth_location.get_itrs(time).cartesian
        itrs_topo = astropy.coordinates.ITRS(topo_itrs_repr, obstime = time, location = earth_location)
        aa = itrs_topo.transform_to(
            astropy.coordinates.AltAz(obstime = time, 
                                      location = earth_location,
                                      pressure = 0)
                                      )
        
        heights = itrs_geo.earth_location.geodetic.height.value

        return aa.alt.degree, aa.az.degree, heights
    
    def plot_teme(self, ax, time):
        """
        Plots satellite constellation in TEME coordinate frame

        :param ax: Axis for plotting onto (must be 3D projection)
        :type ax: :class:`matplotlib.pyplot.Axes`
        :param time: Time at which to plot constellation
        :type time: :class:`astropy.time.Time`
        """

        x, y, z = self.get_teme_position(time)
        x, y, z = x / 1000, y / 1000, z / 1000

        ax.scatter(x, y, z, s = 1, alpha = 0.5)
        ax.set_box_aspect((np.ptp(x), np.ptp(y), np.ptp(z)))
        ax.xaxis.set_pane_color((1, 1, 1, 0))
        ax.yaxis.set_pane_color((1, 1, 1, 0))
        ax.zaxis.set_pane_color((1, 1, 1, 0))
        ax.set_xlabel("x (km)")
        ax.set_ylabel("y (km)")
        ax.set_zlabel("z (km)")
    
    def plot_hcs(self, ax, time, location):
        """
        Plots satellite constellation in HCS coordinate frame

        :param ax: Axis for plotting onto (must be polar projection)
        :type ax: :class:`matplotlib.pyplot.Axes`
        :param time: Time at which to plot constellation
        :type time: :class:`astropy.time.Time`
        :param location: Location on Earth at which to plot constellation
        :type location: :class:`astropy.coordinates.EarthLocation`
        """

        altitudes, azimuths, _ = self.get_hcs_position(time, location)
        ax.plot( np.deg2rad(azimuths), 90 - altitudes, '.')
        ax.set_rmax(90)
        ax.set_yticklabels([])
        ax.set_theta_zero_location('N')
        ax.set_rticks([10, 20, 30, 40, 50, 60, 70, 80, 90])
        ax.set_xticks(np.deg2rad([0, 90, 180, 270]))
        ax.set_xticklabels(['N', 'E', 'S', 'W'])

class ConstellationFromTLE(Constellation):
    """
    This class builds a constellation from a TLE file

    :param tle_file_path: Path to a TLE text file
    :type tle_file_path: str
    """

    def __init__(self, tle_file_path):
        """
        Constructor Method
        """
        
        with open(tle_file_path, 'r') as file:
            lines = file.read().splitlines()

        TLES = [ (l0, l1, l2) for l0, l1, l2 in zip(lines[::3], lines[1::3], lines[2::3]) ]
    
        satellite_ids = []
        constellation = []

        for tle in TLES:
            satellite_ids.append( tle[0].strip() )
            constellation.append(
                sgp4.api.Satrec.twoline2rv(tle[1], tle[2])
                )
        
        self.constellation = sgp4.api.SatrecArray(constellation)
        self.satellite_ids = tuple(satellite_ids)
        self.constellation_size = len( self.satellite_ids )


""" Simple mathematical calculations that are used throughout Lumos """

import numpy as np

def Rx(theta, x, y, z):
    """
    Rotates points :math:`(x, y, z)` about x-axis by angle :math:`\\theta`.
    
    :param theta:
    :type theta: :class:`np.ndarray` or float
    :param x:
    :type x: :class:`np.ndarray` or float
    :param y:
    :type y: :class:`np.ndarray` or float
    :param z:
    :type z: :class:`np.ndarray` or float
    :return: Rotated points :math:`(x^{\\prime}, y^{\\prime}, z^{\\prime})`
    """
    xp = x
    yp = np.cos(theta) * y - np.sin(theta) * z
    zp = np.sin(theta) * y + np.cos(theta) * z
    return xp, yp, zp

def Ry(theta, x, y, z):
    """
    Rotates points :math:`(x, y, z)` about y-axis by angle :math:`\\theta`.
    
    :param theta:
    :type theta: :class:`np.ndarray` or float
    :param x:
    :type x: :class:`np.ndarray` or float
    :param y:
    :type y: :class:`np.ndarray` or float
    :param z:
    :type z: :class:`np.ndarray` or float
    :return: Rotated points :math:`(x^{\\prime}, y^{\\prime}, z^{\\prime})`
    """
    xp = np.cos(theta) * x + np.sin(theta) * z
    yp = y
    zp = -np.sin(theta) * x + np.cos(theta) * z
    return xp, yp, zp

def Rz(theta, x, y, z):
    """
    Rotates points :math:`(x, y, z)` about z-axis by angle :math:`\\theta`.
    
    :param theta:
    :type theta: :class:`np.ndarray` or float
    :param x:
    :type x: :class:`np.ndarray` or float
    :param y:
    :type y: :class:`np.ndarray` or float
    :param z:
    :type z: :class:`np.ndarray` or float
    :return: Rotated points :math:`(x^{\\prime}, y^{\\prime}, z^{\\prime})`
    """
    xp = np.cos(theta) * x - np.sin(theta) * y
    yp = np.sin(theta) * x + np.cos(theta) * y
    zp = z
    return xp, yp, zp

def det_2(a11, a12,
          a21, a22):
    """
    Calculates determinate of 2x2 matrix. Matrix components can be of type
    :class:`np.ndarray` or float.
    
    Returns :math:`\\det \\begin{pmatrix} a_{11} & a_{12} \\\\ a_{21} & a_{22}\\end{pmatrix}`
    """
    return a11 * a22 - a21 * a12

def det_3(a11, a12, a13,
          a21, a22, a23,
          a31, a32, a33):
    """
    Calculates determinate of 3x3 matrix. Matrix components can be of type
    :class:`np.ndarray` or float.

    Returns :math:`\\det \\begin{pmatrix} a_{11} & a_{12} & a_{13} \\\\ a_{21} & a_{22} & a_{23} \\\\ a_{31} & a_{32} & a_{33} \\end{pmatrix}`
    """
    return a11 * det_2(a22, a23, a32, a33) - a12 * det_2(a21, a23, a31, a33) + a13 * det_2(a21, a22, a31, a32)

def inv_3(a11, a12, a13,
          a21, a22, a23,
          a31, a32, a33):
    """
    Calculates inverse of 3x3 matrix. Matrix components can be of type
    :class:`np.ndarray` or float.

    Returns :math:`\\begin{pmatrix} a_{11} & a_{12} & a_{13} \\\\ a_{21} & a_{22} & a_{23} \\\\ a_{31} & a_{32} & a_{33} \\end{pmatrix}^{-1}`
    """
    c = 1 / det_3(a11, a12, a13,
                  a21, a22, a23,
                  a31, a32, a33)

    I11, I12, I13 = c * det_2(a22, a23, a32, a33), c * det_2(a13, a12, a33, a32), c * det_2(a12, a13, a22, a23)
    I21, I22, I23 = c * det_2(a23, a21, a33, a31), c * det_2(a11, a13, a31, a33), c * det_2(a13, a11, a23, a21)
    I31, I32, I33 = c * det_2(a21, a22, a31, a32), c * det_2(a12, a11, a32, a31), c * det_2(a11, a12, a21, a22)

    return I11, I12, I13, \
           I21, I22, I23, \
           I31, I32, I33


"""
Fundamental constants used throughout Lumos.
"""

#: :math:`3.0 \times 10^{8}`  :math:`\frac{m}{s}`
SPEED_OF_LIGHT = 3.0 * 1e8

#: :math:`532 \times 10^{-9}`  :math:`m`
WAVELENGTH = 532 * 1e-9

#: :math:`6378 \times 10^{3}`  :math:`m`
EARTH_RADIUS = 6378 * 1000

#: :math:`1360`  :math:`\frac{W}{m^2}`
SUN_INTENSITY = 1360


"""
Tools for getting TLEs from Celestrak and SpaceTrack
"""

import requests

def query_celestrak(filename, query, value, format, supplemental = False):
    """
    Tool to interface with Celestrak API.
    See `celestrak.org <https://celestrak.org/NORAD/documentation/gp-data-formats.php>`_ for more details.

    :param filename: File where downloaded data is written
    :type filename: str
    :param query: CATNR or INTDES or GROUP or NAME or SPECIAL
    :type query: str
    :param value: Determines how the query is filtered
    :type value: str
    :param format: TLE or 2LE or XML or KVN or JSON or JSON-PRETTY or CSV
    :type format: str
    :param supplemental: Whether to pull from Celestrak's suplemental GP data
    :type supplemental: bool, optional

    """

    url = (f"https://celestrak.org/NORAD/elements/"
           f"{'supplemental/sup-' if supplemental else ''}"
           f"gp.php?{query}={value}&FORMAT={format}")
    
    response = requests.get(url)

    with open(filename, "wb") as file:
        file.write(response.content)


"""
Main brightness calculator for Lumos
"""

import numpy as np
import lumos.constants
import lumos.geometry
import astropy.coordinates

def get_earthshine_panels(sat_z, angle_past_terminator, density):
    """
    Creates a mesh of pixels on Earth's surface which are visible to the satellite and illuminated
    by the sun.

    :param sat_z: The height of the satellite above the center of Earth (meters)
    :type sat_z: float
    :param angle_past_terminator: The angle of the satellite past the terminator (radians)
    :type angle_past_terminator: float
    :param density: The density of the pixels. Grid will have size density x density.
    :type density: int
    :returns: 
        - (x, y, z) - Positions of pixels (meters)
        - (nx, ny, nz) - Normal vectors of pixels
        - areas - Areas of pixels (:math:`m^2`)
    """

    R = lumos.constants.EARTH_RADIUS

    max_angle = np.arccos(R / sat_z)

    angles_off_plane = np.linspace(-max_angle, max_angle, density)
    angles_on_plane = np.linspace(angle_past_terminator, max_angle, density)

    d_phi = abs( angles_off_plane[1] - angles_off_plane[0] )
    d_theta = abs( angles_on_plane[1] - angles_on_plane[0] )

    angles_on_plane, angles_off_plane = np.meshgrid(angles_on_plane, angles_off_plane)
    angles_on_plane, angles_off_plane = angles_on_plane.flatten(), angles_off_plane.flatten()

    # Set up panel positions
    nz = 1 / np.sqrt( 1 + np.tan(angles_on_plane)**2 + np.tan(angles_off_plane)**2 )
    nx = np.tan(angles_off_plane) * nz
    ny = np.tan(angles_on_plane) * nz

    # Clip the panels which aren't visible to the satellite
    visible_to_sat = np.arccos(nz) < max_angle
    angles_off_plane, angles_on_plane = angles_off_plane[visible_to_sat], angles_on_plane[visible_to_sat]
    nx, ny, nz = nx[visible_to_sat], ny[visible_to_sat], nz[visible_to_sat]

    # Calculate Jacobian determinant to get panel areas
    x, y, z = nx * R, ny * R, nz * R

    phi = angles_off_plane
    theta = angles_on_plane
        
    dx_dr = nx / nz * z / R
    dx_dphi = z**3 / (R**2 * np.cos(phi)**2 * np.cos(theta)**2)
    dx_dtheta = - (ny / nz * nx / nz* z**3 ) / (R**2 * np.cos(theta)**2 )
    
    dy_dr = np.tan(theta) * z / R
    dy_dphi = - ( ny / nz * nx / nz * z**3 ) / (R**2 * np.cos(phi)**2 )
    dy_dtheta = dx_dphi
    
    dz_dr = z / R
    dz_dphi = - (nx / nz * z**3 ) / (R**2 * np.cos(phi)**2)
    dz_dtheta = - (ny / nz * z**3 ) / (R**2 * np.cos(theta)**2)

    determinant = (
        dx_dr * (dy_dphi * dz_dtheta - dy_dtheta * dz_dphi) -
        dy_dr * (dx_dphi * dz_dtheta - dx_dtheta * dz_dphi) +
        dz_dr * (dx_dphi * dy_dtheta - dx_dtheta * dy_dphi) )
    
    areas = determinant * d_phi * d_theta
    
    return x, y, z, nx, ny, nz, areas

def get_intensity_satellite_frame(
    sat_surfaces, 
    sat_height, 
    angle_past_terminator,
    observer_position,
    include_sun = True, 
    include_earthshine = True,
    earth_panel_density = 150, 
    earth_brdf = None):
    '''
    Calculates flux scattered by a satellite and seen by an observer.

    :param sat_surfaces: List of surfaces on satellite
    :type sat_surfaces: list[lumos.geometry.Surface]
    :param sat_height: Height of satellite above geodetic nadir (meters)
    :type sat_height: float
    :param angle_past_terminator: Angle of satellite past terminator (radians)
    :type angle_past_terminator: float
    :param observer_position: Position of an observer, measured in brightness frame (meters)
    :type observer_position: :class:`np.ndarray`
    :param include_sun: Whether to include flux scattered by the satellite from the sun
    :type include_sun: bool, optional
    :param include_earthshine: Whether to include flux scattered by the satellite from earthshine
    :type include_earthshine: bool, optional
    :param earth_panel_density: There will be earth_panel_density x earth_panel_density panels in the earthshine mesh
    :type earth_panel_density: int, optional
    :param earth_brdf: A function representing the BRDF of Earth's surface
    :type earth_brdf: function
    :return: Flux of light incident on the observer (W / m^2)
    :rtype: float
    '''
    
    horizon_angle = np.arccos(lumos.constants.EARTH_RADIUS / (lumos.constants.EARTH_RADIUS + sat_height))
    if angle_past_terminator > horizon_angle:
        # Inside earth's shadow
        return 0
    
    observer_x, observer_y, observer_z = observer_position[0], observer_position[1], observer_position[2]

    if np.arccos(observer_z / lumos.constants.EARTH_RADIUS) > horizon_angle + np.deg2rad(1):
        # Not visible to observer
        # Fudging this by 1 degree makes plots much nicer on edges
        return 0

    vector_2_sun = (0, np.cos(angle_past_terminator), - np.sin(angle_past_terminator))

    sat_z = sat_height + lumos.constants.EARTH_RADIUS

    # Distances from observers to satellite
    dist_sat_2_obs = np.sqrt( observer_x**2 + observer_y**2 
                            + (observer_z - sat_z)**2 )
    
    # Unit vectors from observers to satellite
    sat_obs_x = observer_x / dist_sat_2_obs
    sat_obs_y = observer_y / dist_sat_2_obs
    sat_obs_z = (observer_z - sat_z) / dist_sat_2_obs

    if include_earthshine:
        panel_x, panel_y, panel_z, panel_nx, panel_ny, panel_nz, panel_areas = \
            get_earthshine_panels(sat_z, angle_past_terminator, earth_panel_density)
        
        # Distances from earthshine panels to satellite
        dist_panels_2_sat = np.sqrt(panel_x**2 + panel_y**2 + (sat_z - panel_z)**2 )

        panel_sat_x = - panel_x / dist_panels_2_sat
        panel_sat_y = - panel_y / dist_panels_2_sat
        panel_sat_z = (sat_z - panel_z) / dist_panels_2_sat

        # Panel Normalization
        panel_normalizations = (panel_ny * vector_2_sun[1] + panel_nz * vector_2_sun[2])

        panel_brdfs = earth_brdf(vector_2_sun,
                                (panel_nx, panel_ny, panel_nz),
                                (panel_sat_x, panel_sat_y, panel_sat_z))

    intensity = 0

    for surface in sat_surfaces:
        
        if callable(surface.normal):
            surface_normal = surface.normal(angle_past_terminator)
        else:
            surface_normal = surface.normal
        
        sun_contribution = 0
        earth_contribution = 0
        
        surface_normalization = surface_normal[1] * vector_2_sun[1] + surface_normal[2] * vector_2_sun[2]
        surface_normalization = np.clip(surface_normalization, 0, None)

        observer_normalization = surface_normal[0] * sat_obs_x + surface_normal[1] * sat_obs_y + surface_normal[2] * sat_obs_z
        observer_normalization = np.clip(observer_normalization, 0, None)

        if include_sun:
            surface_brdf = surface.brdf(vector_2_sun,
                                        surface_normal,
                                        (sat_obs_x, sat_obs_y, sat_obs_z))

            sun_contribution = sun_contribution + surface_brdf * surface_normalization * observer_normalization

        if include_earthshine:

            surface_normalizations = np.clip(
                - surface_normal[0] * panel_sat_x \
                - surface_normal[1] * panel_sat_y \
                - surface_normal[2] * panel_sat_z,
                0,
                None
            )

            panel_observing_normalization = np.clip(
                panel_nx * panel_sat_x
                + panel_ny * panel_sat_y
                + panel_nz * panel_sat_z,
                0,
                None
            )

            surface_brdf = surface.brdf( (-panel_sat_x, -panel_sat_y, -panel_sat_z),
                                          surface_normal,
                                         (sat_obs_x, sat_obs_y, sat_obs_z) )
            
            earth_contribution = earth_contribution \
                  + np.sum(panel_brdfs * surface_brdf 
                         * panel_normalizations * observer_normalization 
                         * surface_normalizations * panel_observing_normalization
                         * panel_areas / dist_panels_2_sat**2 )
        
        intensity = intensity + lumos.constants.SUN_INTENSITY * surface.area * (sun_contribution + earth_contribution) / dist_sat_2_obs ** 2

    return intensity

def get_brightness_coords(
    sat_alt, 
    sat_az, 
    sat_height, 
    sun_alt, 
    sun_az):
    """
    Converts from HCS observer coordinates to brightness coordinates.

    :param sat_alt: Altitude of satellite (degrees)
    :type sat_alt: float or :class:`np.ndarray`
    :param sat_az: Altitude of satellite (degrees)
    :type sat_az: float or :class:`np.ndarray`
    :param sat_height: Geodetic height of satellite (meters)
    :type sat_height: float or :class:`np.ndarray`
    :param sun_alt: Altitude of sun (degrees)
    :type sun_alt: float
    :param sun_az: Azimuth angle of sun (degrees)
    :type sun_az: float
    :returns:
        - (obs_x, obs_y, obs_z) - Vector from satellite to observer in brightness frame (meters)
        - angle_past_terminator - Angle of satellite past terminator (radians)
    """

    sat_x, sat_y, sat_z = lumos.conversions.altaz_to_unit(sat_alt, sat_az)
    sun_x, sun_y, sun_z = lumos.conversions.altaz_to_unit(sun_alt, sun_az)

    phi = np.arccos(sat_z) - np.arcsin(lumos.constants.EARTH_RADIUS / (lumos.constants.EARTH_RADIUS + sat_height) * np.sqrt(sat_x**2 + sat_y**2) )
    theta = np.arctan2(sat_y, sat_x)

    Zx = np.sin(phi) * np.cos(theta)
    Zy = np.sin(phi) * np.sin(theta)
    Zz = np.cos(phi)

    dot = Zx * sun_x + Zy * sun_y + Zz * sun_z
    beta = 1 / np.sqrt(1 - dot**2)
    alpha = - beta * dot

    Yx = alpha * Zx + beta * sun_x
    Yy = alpha * Zy + beta * sun_y
    Yz = alpha * Zz + beta * sun_z

    flip = np.sign(Yx * sun_x + Yy * sun_y + Yz * sun_z)
    Yx, Yy, Yz = flip * Yx, flip * Yy, flip * Yz

    Xx = Yy * Zz - Zy * Yz
    Xy = Zx * Yz - Yx * Zz
    Xz = Yx * Zy - Zx * Yy

    T11, T12, T13, \
    T21, T22, T23, \
    T31, T32, T33  = lumos.functions.inv_3(Xx, Yx, Zx, 
                                           Xy, Yy, Zy, 
                                           Xz, Yz, Zz)
    
    dist_to_sat = np.sqrt(lumos.constants.EARTH_RADIUS**2 
    + (lumos.constants.EARTH_RADIUS + sat_height)**2 
    - 2 * lumos.constants.EARTH_RADIUS * (lumos.constants.EARTH_RADIUS + sat_height) * np.cos(phi))

    obs_x = - dist_to_sat * (T11 * sat_x + T12 * sat_y + T13 * sat_z)
    obs_y = - dist_to_sat * (T21 * sat_x + T22 * sat_y + T23 * sat_z)
    obs_z = (lumos.constants.EARTH_RADIUS + sat_height) - dist_to_sat * (T31 * sat_x + T32 * sat_y + T33 * sat_z)

    angle_past_terminator = -np.arcsin(T31 * sun_x + T32 * sun_y + T33 * sun_z)

    return obs_x, obs_y, obs_z, angle_past_terminator

def get_intensity_observer_frame(
    sat_surfaces, 
    sat_heights, 
    sat_altitudes, 
    sat_azimuths,
    sun_altitude,
    sun_azimuth,
    include_sun = True,
    include_earthshine = True,
    earth_panel_density = 150,
    earth_brdf = None):

    """
    Calculates the flux of a satellite seen by an observer.

    :param sat_surfaces: List of surfaces on satellite
    :type sat_surfaces: list[lumos.geometry.Surface]
    :param sat_heights: Heights of satellite above geodetic nadir (meters)
    :type sat_heights: float or :class:`np.ndarray`
    :param sat_altitudes: Satellite altitudes in HCS frame (degrees)
    :type sat_altitudes: float or :class:`np.ndarray`
    :param sat_azimuths: Satellite azimuths in HCS frame (degrees)
    :type sat_azimuths: float or :class:`np.ndarray`
    :param sun_altitude: Altitude of sun in HCS frame (degrees)
    :type sun_altitude: float
    :param sun_azimuth: Azimuth of sun in HCS frame (degrees)
    :type sun_azimuth: float
    :param include_sun: Whether to include flux scattered by the satellite from the sun
    :type include_sun: bool, optional
    :param include_earthshine: Whether to include flux scattered by the satellite from earthshine
    :type include_earthshine: bool, optional
    :param earth_panel_density: There will be earth_panel_density x earth_panel_density panels in the earthshine mesh
    :type earth_panel_density: int, optional
    :param earth_brdf: A function representing the BRDF of Earth's surface
    :type earth_brdf: function
    :return: Flux of light incident on the observer (W / m^2)
    :rtype: float or :class:`np.ndarray`
    """
    
    if sun_altitude > 0:
        raise ValueError(f"Observatory is in daylight! Sun Altitude = {sun_altitude}")
    
    sat_alts, sat_azs = np.asarray(sat_altitudes), np.asarray(sat_azimuths)
    output_shape = sat_alts.shape

    if not isinstance(sat_heights, np.ndarray):
        sat_heights = sat_heights * np.ones(output_shape)

    sat_heights, sat_alts, sat_azs = sat_heights.flatten(), sat_alts.flatten(), sat_azs.flatten()
    intensity = np.zeros_like(sat_alts)

    obs_x, obs_y, obs_z, angle_past_terminator = get_brightness_coords(sat_alts, sat_azs, sat_heights, 
        sun_altitude, sun_azimuth)

    for (i, (x, y, z, alpha, h)) in enumerate(zip(obs_x, obs_y, obs_z, angle_past_terminator, sat_heights)):
        intensity[i] = get_intensity_satellite_frame(
            sat_surfaces, h, alpha, (x, y, z),
            include_sun, include_earthshine, earth_panel_density, earth_brdf)

    intensity = intensity.reshape(output_shape)
    return intensity

def get_sun_alt_az(time, observer_location):
    
    """
    Convenience function for finding the altitude and azimuth of the sun

    :param time: Time of observation
    :type time: :class:`astropy.time.Time`
    :param observer_location: Location of observation
    :type observer_location: :class:`astropy.coordinates.EarthLocation`
    """

    aa_frame = astropy.coordinates.AltAz(obstime = time, location = observer_location)
    sun_altaz = astropy.coordinates.get_sun(time).transform_to(aa_frame)
    return sun_altaz.alt.degree, sun_altaz.az.degree




""" Helpful plotting functions for Lumos """

import matplotlib.pyplot as plt
import matplotlib.patches
import matplotlib.cm

import numpy as np

import lumos.conversions
import lumos.constants
import lumos.geometry

import os
import cv2

def BRDF_1D(ax, brdf_model, incident_angles = (10, 30, 50, 70), log_space = True):
    """
    Plots the in-plane BRDF at given incident angles.

    :param ax: Plotting axis
    :type ax: :class:`matplotlib.pyplot.axes`
    :param brdf_model: BRDF function to plot
    :type brdf_model: function
    :param incident_angles: Incident angles for which to plot BRDF model (degrees)
    :type incident_angles: tuple, optional
    :param log_space: Whether to plot in log space
    :type log_space: bool, optional
    """
    for angle in incident_angles:
        x = np.linspace(-90, 90, 400)
        ix, iy, iz = lumos.conversions.spherical_to_unit(np.deg2rad(angle), np.pi)
        ox, oy, oz = lumos.conversions.spherical_to_unit(np.deg2rad(x), 0)
        y = brdf_model((ix, iy, iz), (0, 0, 1), (ox, oy, oz))
        label = r"$\phi_{in}$ = " + f"{angle:0.1f}°"
        if log_space:
            ax.semilogy(x, y, label = label)
        else:
            ax.plot(x, y, label = label)
    ax.legend()

def BRDF_2D(polar_ax, brdf_model, incident_angle):
    """
    Plots the BRDF at given incident angle.

    :param ax: Plotting axis. Must have polar projection.
    :type ax: :class:`matplotlib.axes.Axes`
    :param brdf_model: BRDF function to plot
    :type brdf_model: function
    :param incident_angle: Incident angle for which to plot BRDF model (degrees)
    :type incident_angle: float
    """
    phi_out = np.linspace(0, 90, 180)
    theta_out = np.linspace(0, 360, 360)
    phi_out, theta_out = np.meshgrid(phi_out, theta_out)

    ix, iy, iz = lumos.conversions.spherical_to_unit(np.deg2rad(incident_angle), np.pi)
    ox, oy, oz = lumos.conversions.spherical_to_unit(np.deg2rad(phi_out), np.deg2rad(theta_out))
    y = brdf_model((ix, iy, iz), (0, 0, 1), (ox, oy, oz))
    polar_ax.contourf(np.deg2rad(theta_out), phi_out, np.log10(y))
    polar_ax.set_title(r"$\phi_{in}$ = " + f"{incident_angle:0.1f}°")
    polar_ax.set_xticks([])
    polar_ax.set_yticks([])
    polar_ax.set_theta_zero_location('N')

def contour_satellite_frame(
    ax,
    observers,
    values,
    cmap = 'plasma',
    levels = None
    ):

    """
    Makes a contour plot

    :param ax: Plotting axis
    :type ax: :class:`matplotlib.pyplot.axes`
    :param observers: Mesh of observers
    :type observers: :class:`lumos.geometry.GroundObservers`
    :param values: 2D array of values to plot
    :type values: :class:`np.ndarray`
    :param cmap: Matplotlib colormap to make contour with
    :type cmap: str, optional
    :param levels: Minimum and maximum value to plot
    :type levels: tuple, optional
    """

    if levels is None:
        levels = (values.min(), values.max())

    max_dist = observers.max_angle * lumos.constants.EARTH_RADIUS / 1000

    ax.set_xlim(-max_dist, max_dist)
    ax.set_ylim(-max_dist, max_dist)
    ax.set_aspect("equal")
    
    ax.set_xlabel("Distance on Plane (km)")
    ax.set_ylabel("Distance off Plane (km)")

    circ = matplotlib.patches.Circle((0, 0), max_dist , transform=ax.transData, facecolor = (1, 1, 1))
    ax.add_patch(circ)

    cs = ax.contourf(observers.dists_on_plane / 1000,
                    observers.dists_off_plane / 1000,
                    values,
                    cmap = matplotlib.colormaps[cmap],
                    norm = matplotlib.colors.Normalize(levels[0], levels[1]),
                    levels = range(levels[0], levels[1] + 1),
                    extend = 'both')
    
    for collection in cs.collections:
        collection.set_clip_path(circ)

def mark_angles_above_horizon_satellite_frame(
    ax,
    observers,
    angles_above_horizon = (15, 30) 
    ):

    """
    Marks discrete angles above horizon using dashed circles

    :param ax: Matplotlib axis for plotting
    :type ax: :class:`matplotlib.pyplot.axes`
    :param observers: Mesh of observers
    :type observers: :class:`lumos.geometry.GroundObservers`
    :param angles_above_horizon: Angles to mark (degrees)
    :type angles_above_horizon: tuple, optional
    """

    theta = np.linspace(0, 2 * np.pi, 50)

    for angle in angles_above_horizon:
        angle = np.deg2rad(angle)
        R = (lumos.constants.EARTH_RADIUS / 1000
        * (np.pi/2 - angle - np.arcsin(np.cos(observers.max_angle) * np.cos(angle)))
        )

        x = R * np.cos(theta)
        y = R * np.sin(theta)

        annotation_loc = (0.707 * R, 0.707 * R)
        y = np.where( (x - annotation_loc[0])**2 + (y - annotation_loc[1]) ** 2 < 100**2, np.inf, y)

        ax.plot(x, y, '--k')
        ax.annotate(f'{np.rad2deg(angle):.0f}°', annotation_loc, fontsize = 10, color = 'black', 
                horizontalalignment = 'center', verticalalignment = 'center')

def contour_observer_frame(
    ax,
    altitudes,
    azimuths,
    values,
    levels = None,
    cmap = 'plasma'
    ):

    """
    Creates contour plot in observer frame

    :param ax: Matplotlib axis for plotting on
    :type ax: :class:`matplotlib.pyplot.axes`
    :param altitudes: Altitudes in HCS frame (degrees)
    :type altitudes: :class:`np.ndarray`
    :param azimuths: Azimuths in HCS frame (degrees)
    :type azimuths: :class:`np.ndarray`
    :param values: Values to plot
    :type values: :class:`np.ndarray`
    :param levels: Minimum and maximum value to plot
    :type levels: tuple, optional
    :param cmap: Matplotlib colormap to use
    :type cmap: str
    """

    if levels is None:
        levels = (values.min(), values.max())

    ax.contourf(
        np.deg2rad(azimuths),
        90 - altitudes,
        values,
        cmap = matplotlib.colormaps[cmap],
        norm = matplotlib.colors.Normalize(levels[0], levels[1]),
        levels = range(levels[0], levels[1] + 1),
        extend = 'both'
        )

    ax.set_rmax(90)
    ax.set_yticklabels([])
    ax.set_theta_zero_location('N')
    ax.set_rticks([10, 20, 30, 40, 50, 60, 70, 80, 90])
    ax.set_xticks(np.deg2rad([0, 90, 180, 270]))
    ax.set_xticklabels(['N', 'E', 'S', 'W'])
    ax.set_rlabel_position(-22.5)
    ax.grid(True)

def colorbar(cax, levels):
    cmap = matplotlib.colormaps['plasma_r']
    norm = matplotlib.colors.Normalize(levels[0], levels[1])
    plt.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap), cax = cax, extend = 'both')
    cax.set_aspect(3)

def mark_sun_azimuth_observer_frame(
    ax,
    sun_azimuth,
    ):
    
    """
    Adds small sun on edge of plot

    :param ax: Axis for plotting
    :type ax: :class:`matplotlib.pyplot.axes`
    :param sun_azimuth: Azimuth of sun (degrees)
    :type sun_azimuth: float
    """
    ax.plot([np.deg2rad(sun_azimuth)], [95], 
            marker = (3, 0, 180 - sun_azimuth), 
            markersize = 6, 
            color = "white", 
            clip_on = False)
    
    ax.plot([np.deg2rad(sun_azimuth)], [101], 
            marker = '$\u2600$', 
            markersize = 10, 
            color = "orange", 
            clip_on = False)

def mark_sun_altitude_observer_frame(
    cax,
    sun_altitude
    ):

    """
    Adds colorbar with small sun to mark time of evening

    :param cax: Matplotlib axis for plotting
    :type cax: :class:`matplotlib.pyplot.axes`
    :param sun_altitude: Altitude of sun (degrees)
    :type sun_altitude: float
    """
    
    color_list = ('#15171c', '#20242d', '#25406a', '#4872bc', '#88a5d1', '#b5c9e6')
    evening_cmap = matplotlib.colors.LinearSegmentedColormap.from_list('Evening', color_list)
    norm = matplotlib.colors.Normalize(vmin = -18, vmax = 0)

    plt.colorbar(matplotlib.cm.ScalarMappable(norm = norm, cmap = evening_cmap), cax = cax)
    cax.set_aspect('equal')
    cax.set_xlim(0, 1)
    cax.yaxis.set_tick_params(which = 'minor', length = 0, rotation = 0, pad = 35)
    cax.yaxis.set_tick_params(which = 'major', length = 5, pad = 5)
    cax.xaxis.set_label_position("top")
    cax.set_yticks(ticks = [-18, -12, -6, 0], labels = ['Night', '', '', 'Day'], fontsize = 12)
    cax.set_yticks(ticks = [-15.5, -9.5, -3.5], labels = ['Astronomical\nTwilight', 'Nautical\nTwilight', 'Civil\nTwilight'], minor = True, fontsize = 8, ma = 'center', ha = 'center')
    plot_alt = np.clip(sun_altitude, -18, 0)
    cax.plot([-0.4], [plot_alt], marker = (3, 0, 270), color = 'white', markersize = 6, clip_on = False)
    cax.plot([-1.0], [plot_alt], marker = '$\u2600$', color = 'orange', markersize = 10, clip_on = False)

def create_video(image_folder_path, video_output_path, frame_rate):
    """
    Combines folder of .png images to create a .mp4 video

    :param image_folder_path: Path to folder containing images.
    :type image_folder_path: str
    :param video_output_path: Destination path for output video
    :type video_output_path: str
    :param frame_rate: Video frames per second
    :type frame_rate: int
    """
    
    images = [img for img in sorted(os.listdir(image_folder_path)) if img.endswith(".png")]
    frame = cv2.imread(os.path.join(image_folder_path, images[0]))
    
    height, width, _ = frame.shape
    
    video = cv2.VideoWriter(video_output_path, 0, frame_rate, (width, height))
    
    for image in images:
        video.write(cv2.imread(os.path.join(image_folder_path, image)))
    
    cv2.destroyAllWindows()
    video.release()


""" Conversions which are used throughout Lumos """

import numpy as np
import lumos.constants
import lumos.functions

def intensity_to_ab_mag(intensity, clip = True):
    """
    Converts from intensity to AB Magnitude.
    If clip is set to True, outputs below 12 AB Magnitude will be clipped.

    :param intensity: Intensity :math:`\\frac{W}{m^2}`
    :type intensity: :class:`np.ndarray` or float
    :param clip: Whether or not to clip very small intensities
    :type clip: bool, optional
    :return: AB Magnitude
    :rtype: :class:`np.ndarray` or float
    """
    log_val = intensity * lumos.constants.WAVELENGTH / (lumos.constants.SPEED_OF_LIGHT * 3631e-26)
    if clip:
        log_val = np.clip(log_val, 10e-6, None)
    ab_mag = -2.5 * np.log10( log_val )
    return ab_mag

def altaz_to_unit(altitude, azimuth):
    """
    Converts altitude and azimuth to a unit vector.

    :param altitude: Altitude in HCS frame (degrees)
    :type altitude: :class:`np.ndarray` or float
    :param azimuth: Azimuth in HCS frame (degrees)
    :type azimuth: :class:`np.ndarray` or float
    :return: Unit vector components :math:`(x, y, z)`
    :rtype: tuple
    """
    phi = np.deg2rad(90 - altitude)
    theta = np.deg2rad(azimuth)
    x = np.sin(phi) * np.cos(theta)
    y = np.sin(phi) * np.sin(theta)
    z = np.cos(phi)
    return x, y, z

def unit_to_spherical(x, y, z):
    """
    Converts a unit vector :math:`(x, y, z)` to spherical coordinates
    :math:`(\\phi, \\theta)`

    :param x:
    :type x: :class:`np.ndarray` or float
    :param y:
    :type y: :class:`np.ndarray` or float
    :param z:
    :type z: :class:`np.ndarray` or float
    :return: :math:`(\\phi, \\theta)`
    :rtype: tuple
    """
    phi = np.arccos(z)
    theta = np.arctan2(y, x)
    phi = np.rad2deg(phi)
    theta = np.rad2deg(theta)
    return phi, theta

def spherical_to_unit(phi, theta):
    """
    Converts from spherical coordinates :math:`(\\phi, \\theta)`
    to a unit vector :math:`(x, y, z)`

    :param phi:
    :type phi: :class:`np.ndarray` or float
    :param theta:
    :type theta: :class:`np.ndarray` or float
    :return: :math:`(x, y, z)`
    :rtype: tuple
    """
    x = np.sin(phi) * np.cos(theta)
    y = np.sin(phi) * np.sin(theta)
    z = np.cos(phi)
    return x, y, z


"""
Geometry objects which are used throughout Lumos
"""

import numpy as np
import lumos.constants
class Surface:
    """
    Container to hold area, normal vector, and BRDF of a surface 

    :param area: Area of surface :math:`m^2`
    :type area: float
    :param normal: Normal vector of surface. Measured in brightness frame. A function of
    the angle past terminator may be passed, which must return the surface normal as 
    a :class:`np.ndarray` for given angles past the terminator.
    :type normal: :class:`np.ndarray` or function
    :param brdf: Bidirectional Reflectance Distribution Function (BRDF) of surface.
    :type brdf: function
    """
    def __init__(self, area, normal, brdf):
        """
        Constructor Method
        """
        self.area = area
        self.normal = normal
        self.brdf = brdf
    
    def __str__(self):
        """
        Printing Method
        """
        return f'| Surface \n' \
               f'|-- Area: {self.area:.2f} m^2 \n' \
               f'|-- Normal Vector: ' \
               f'<{self.normal[0]:.2f}, {self.normal[1]:.2f}, {self.normal[2]:.2f}> \n'

class EarthMesh:
    """
    A mesh of points on Earth's surface

    :param angles_off_plane: 1D array of the angles-off-plane of the mesh.
    :type angles_off_plane: :class:`np.ndarray`
    :param angles_on_plane: 1D array of the angles-on-plane of the mesh.
    :type angles_on_plane: :class:`np.ndarray`
    """
    def __init__(self, angles_off_plane, angles_on_plane):
        """
        Constructor method
        """
        self.d_phi = np.abs(angles_off_plane[1] - angles_off_plane[0])
        self.d_theta = np.abs(angles_on_plane[1] - angles_on_plane[0])
        
        self.angles_off_plane, self.angles_on_plane = np.meshgrid(angles_off_plane,
                                                                  angles_on_plane)
        
        self.dists_off_plane = lumos.constants.EARTH_RADIUS * self.angles_off_plane
        self.dists_on_plane = lumos.constants.EARTH_RADIUS * self.angles_on_plane
        
        # Set up panel positions
        self.z = ( lumos.constants.EARTH_RADIUS 
                  / np.sqrt(1 + np.tan(self.angles_on_plane)**2 
                            + np.tan(self.angles_off_plane)**2 ) )
                  
        self.x = np.tan(self.angles_off_plane) * self.z
        self.y = np.tan(self.angles_on_plane) * self.z
        
        # Get normal vectors of panels
        self.nx = self.x / lumos.constants.EARTH_RADIUS
        self.ny = self.y / lumos.constants.EARTH_RADIUS
        self.nz = self.z / lumos.constants.EARTH_RADIUS

        self.shape = self.z.shape
        
class GroundObservers(EarthMesh):
    """
    A mesh of observers visible to the satellite and on the night side of earth

    :param sat_height: Geodetic height of satellite (meters)
    :type sat_height: float
    :param angle_past_terminator: Angle of satellite past the terminator (radians)
    :type angle_past_terminator: float
    :param density: Mesh will have size density x density
    :type density: int
    """

    def __init__(self, sat_height, angle_past_terminator, density):
        """
        Constructor Method
        """
        self.max_angle = np.arccos(lumos.constants.EARTH_RADIUS 
                                   / (lumos.constants.EARTH_RADIUS + sat_height))
        self.min_angle = angle_past_terminator
        
        angles_off_plane = np.linspace(-self.max_angle, self.max_angle, density)
        angles_on_plane = np.linspace(-self.max_angle, self.min_angle, density)
        
        super().__init__(angles_off_plane, angles_on_plane)
    
    def __iter__(self):
        """
        Iteration Method
        """
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                yield i, j, (self.x[i, j], self.y[i, j], self.z[i, j])


""" A catalog of useful and common BRDF models """
import numpy as np

def LAMBERTIAN(albedo):
    """
    Lambertian scattering model.

    :math:`BRDF = \\frac{\\rho}{\\pi}`

    :param albedo: Albedo, should be between 0.0 and 1.0
    :type albedo: float
    :return: BRDF function, where f = BRDF(incident vector, normal_vector, outgoing_vector)
    :rtype: function
    """
    def BRDF(incident_vector, normal_vector, outgoing_vector):
        ix, iy, iz = incident_vector
        nx, ny, nz = normal_vector
        ox, oy, oz = outgoing_vector

        brdf = np.where(ix*nx + iy*ny + iz*nz < 0, 0, albedo / np.pi)
        brdf = np.where(ox*nx + oy*ny + oz*nz < 0, 0, albedo / np.pi)

        return brdf
    
    return BRDF

def ABG(A, B, g):
    """
    ABG scattering model. 
    Described here: `Using the ABG model <https://support.zemax.com/hc/en-us/articles/4408106806163-Fitting-ABg-scattering-coefficients-from-raw-measured-data>`_

    :math:`BRDF = \\frac{A}{B + x^g}`

    :param A: Amplitude parameter, A/B is the specular peak
    :type A: float
    :param B: Knee parameter, B determines where the BRDF transitions to exponential decay
    :type B: float
    :param g: Slope parameter, g determines the slope of the BRDF in log-log space
    :type g: float
    :return: BRDF function, where f = BRDF(incident vector, normal_vector, outgoing_vector)
    :rtype: function
    """
    def BRDF(incident_vector, normal_vector, outgoing_vector):
        ix, iy, iz = incident_vector
        nx, ny, nz = normal_vector
        ox, oy, oz = outgoing_vector
        
        dot = ix*nx + iy*ny + iz*nz
        rx = 2 * dot * nx - ix
        ry = 2 * dot * ny - iy
        rz = 2 * dot * nz - iz

        dot = ox*nx + oy*ny + oz*nz
        beta_x = ox - dot * nx
        beta_y = oy - dot * ny
        beta_z = oz - dot * nz

        dot = rx*nx + ry*ny + rz*nz
        beta0_x = rx - dot * nx
        beta0_y = ry - dot * ny
        beta0_z = rz - dot * nz

        x = np.sqrt( (beta_x - beta0_x)**2 + (beta_y - beta0_y)**2 + (beta_z - beta0_z)**2 )

        brdf = A / (B + x**g)

        brdf = np.where(ix*nx + iy*ny + iz*nz < 0, 0, brdf)
        brdf = np.where(ox*nx + oy*ny + oz*nz < 0, 0, brdf)
        
        return brdf
    
    return BRDF

def GAUSSIAN(A, sigma):
    """
    Gaussian scattering model. 

    :math:`BRDF = A exp(\\frac{\\hat{w}_r\\cdot\\hat{w}_o - 1}{\\sigma})`

    :param A: Height of specular peak of BRDF
    :type A: float
    :param sigma: Width of specular peak
    :type sigma: float
    :return: BRDF function, where f = BRDF(incident vector, normal_vector, outgoing_vector)
    :rtype: function
    """

    def BRDF(incident_vector, normal_vector, outgoing_vector):
        ix, iy, iz = incident_vector
        nx, ny, nz = normal_vector
        ox, oy, oz = outgoing_vector
        
        dot = ix*nx + iy*ny + iz*nz
        rx = 2 * dot * nx - ix
        ry = 2 * dot * ny - iy
        rz = 2 * dot * nz - iz

        dot = rx * ox + ry * oy + rz * oz
        brdf = A * np.exp( (dot - 1) / sigma )

        brdf = np.where(ix*nx + iy*ny + iz*nz < 0, 0, brdf)
        brdf = np.where(ox*nx + oy*ny + oz*nz < 0, 0, brdf)
        
        return brdf
    
    return BRDF

def PHONG(Kd, Ks, n):
    """
    Phong scattering model. 

    :math:`BRDF = \\frac{K_d}{\\pi} + K_s\\frac{n + 2}{2 \\pi} (\\hat{w}_r \\cdot \\hat{w}_o)^n`

    :param Kd: Diffuse coefficient. Ensure :math:`Kd + Ks \\leq 1`
    :type Kd: float
    :param Ks: Specular coefficient. Ensure :math:`Kd + Ks \\leq 1`
    :type Ks: float
    :param n: Controls width of specular peak.
    :type n: float
    :return: BRDF function, where f = BRDF(incident vector, normal_vector, outgoing_vector)
    :rtype: function
    """

    def BRDF(incident_vector, normal_vector, outgoing_vector):
        ix, iy, iz = incident_vector
        nx, ny, nz = normal_vector
        ox, oy, oz = outgoing_vector
        
        dot = ix*nx + iy*ny + iz*nz
        rx = 2 * dot * nx - ix
        ry = 2 * dot * ny - iy
        rz = 2 * dot * nz - iz

        dot = rx*ox + ry*oy + rz*oz
        dot = np.clip(dot, 0, 1)

        brdf = Kd / np.pi + Ks * (n + 2) / (2 * np.pi) * dot**n

        brdf = np.where(ix*nx + iy*ny + iz*nz < 0, 0, brdf)
        brdf = np.where(ox*nx + oy*ny + oz*nz < 0, 0, brdf)
        
        return brdf
    
    return BRDF

def BINOMIAL(B, C, d, l1):
    """
    Binomial scattering model proposed by Greynolds.

    Described here: `Physically Realistic BRDF Models <https://www.researchgate.net/publication/300345613_General_physically-realistic_BRDF_models_for_computing_stray_light_from_arbitrary_isotropic_surfaces>`_

    :param B: b_i model coefficients
    :type B: :class:`np.ndarray`
    :param C: c_i model coefficients
    :type C: :class:`np.ndarray`
    :param d: specularity constant
    :type d: float
    :param l1: Minimum Gaussian index
    :type l1: int
    :return: BRDF function, where f = BRDF(incident vector, normal_vector, outgoing_vector)
    :rtype: function
    """

    def BRDF(incident_vector, normal_vector, outgoing_vector):
        ix, iy, iz = incident_vector
        nx, ny, nz = normal_vector
        ox, oy, oz = outgoing_vector

        dot = ix*nx + iy*ny + iz*nz
        rx = 2 * dot * nx - ix
        ry = 2 * dot * ny - iy
        rz = 2 * dot * nz - iz

        dot = ox*nx + oy*ny + oz*nz
        rho_x = ox - dot * nx
        rho_y = oy - dot * ny
        rho_z = oz - dot * nz

        dot = rx*nx + ry*ny + rz*nz
        rho0_x = rx - dot * nx
        rho0_y = ry - dot * ny
        rho0_z = rz - dot * nz

        D = np.sqrt( (rho_x - rho0_x)**2 + (rho_y - rho0_y)**2 + (rho_z - rho0_z)**2 )
        V = rho_x * rho0_x + rho_y * rho0_y + rho_z * rho0_z

        log_brdf = np.zeros_like(D)
        
        n, m = B.shape
        _, j = C.shape

        for k in range(n):
            term_1 = np.zeros_like(D)

            for i in range(m):
                term_1 = term_1 + B[k, i] * D ** i
            
            term_2 = np.zeros_like(D)
            for i in range(j):
                term_2 = term_2 + C[k, i] * np.log10(1 + d**(i + l1) * D**2)
            
            log_brdf = log_brdf + (term_1 + 0.5 * term_2) * V**k
        
        brdf = 10**log_brdf

        brdf = np.where(ix*nx + iy*ny + iz*nz < 0, 0, brdf)
        brdf = np.where(ox*nx + oy*ny + oz*nz < 0, 0, brdf)
        
        return brdf
    
    return BRDF


"""
Tools for working with BRDF models and data
"""

import numpy as np
import scipy.optimize
import lumos.conversions

def read_brdf(brdf_file_path):
    """
    Reads data from a BRDF file

    :param brdf_file_path: Path to the BRDF file
    :type brdf_file_path: str
    :return: :math:`\\phi_i, \\theta_i, \\phi_o, \\theta_o, BRDF`
    :rtype: tuple[:class:`np.ndarray`]
    """

    data = np.loadtxt(brdf_file_path, skiprows = 1)

    phi_in = data[:, 0]
    theta_in = data[:, 1]
    phi_out = data[:, 2]
    theta_out = data[:, 3]
    brdf = data[:, 4]

    return phi_in, theta_in, phi_out, theta_out, brdf

def fit(
        data_file,
        model_func,
        bounds = None,
        p0 = None,
        log_space = True,
        clip = 0
        ):
    
    """
    Fits a model to experimental data.

    :param data_file: A path to a .csv file containing BRDF data. DESCRIBE FORMAT.
    :type data_file: str
    :param model_func: A BRDF model to fit
    :type model_func: function
    :param bounds: Bounds passed to :function:`scipy.optimize.curve_fit`
    :type bounds: tuple
    :param p0: Initial guess for parameters passed to :function:`scipy.optimize.curve_fit`
    :type p0: tuple
    :param log_space: Whether or not to fit BRDF in log_space
    :type log_space: bool
    :param clip: Removes BRDF data below this cutoff from fitting
    :type clip: float
    """

    phi_in, theta_in, phi_out, theta_out, brdf = read_brdf(data_file)
    phi_in, theta_in = np.deg2rad(phi_in), np.deg2rad(theta_in)
    phi_out, theta_out = np.deg2rad(phi_out), np.deg2rad(theta_out)

    mask = brdf > clip

    phi_in = phi_in[mask]
    theta_in = theta_in[mask]
    phi_out = phi_out[mask]
    theta_out = theta_out[mask]
    brdf = brdf[mask]

    indexes = np.arange(brdf.size)

    ix, iy, iz = lumos.conversions.spherical_to_unit(phi_in, theta_in)
    ox, oy, oz = lumos.conversions.spherical_to_unit(phi_out, theta_out)

    def fit_function(idx, *params):
        model_brdf = model_func(*params)

        idx = idx.astype(int)

        f = model_brdf(
            (ix[idx], iy[idx], iz[idx]),
            (0, 0, 1),
            (ox[idx], oy[idx], oz[idx])
            )
        
        return np.log10(f) if log_space else f

    popt, _ = scipy.optimize.curve_fit(fit_function, 
                                       indexes, 
                                       np.log10(brdf) if log_space else brdf, 
                                       bounds = bounds,
                                       p0 = p0)

    return popt

def pack_binomial_parameters(n, m, l1, l2, *params):
    """
    Convert list into B & C matrices, which can then be passed to the Binomial Model

    :param n: n
    :type n: int
    :param m: m
    :type m: int
    :param l1: l1
    :type l1: int
    :param l2: l2, ensure l2 > l1
    :type l2: int
    :param params: list of values
    :type params: list[float]
    :return: B, C, d
    :rtype: :class:`np.ndarray`, :class:`np.ndarray`, float
    """
    params = np.array(params)
    B = np.reshape( params[:n * m], (n, m) )
    C = np.reshape( params[n * m:], (n, l2 - l1))
    return B, C